/*----------------------------------------------------------------------------
;
; Generate byte pattern to stdout.  (c) 2019 PaulHoule.com
;
; Change Log
; ----------
; 20190901 ver 2.1: allow double quoted string arg(s); also, decimal output
;	size arg can be omitted if first argument is a string or @{file}.
;
;---------------------------------------------------------------------------*/

#ifdef PH_								// if "tiny .exe" stdlib (optional)
#include "ph_.h"
#else									// else normal compiler stdlib
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <stdint.h>
#endif

typedef struct spbuf {					// stores data buffers in a list
	struct spbuf *pNext;				// next PBUF, or NULL if last
	size_t totprevbused;				// sum of all previous bused in list
	uint32_t  bbytes;					// bytes allocated to buf[]
	uint32_t  bused;					// bytes currently used in buf[]
	uint8_t buf[65536];					// (keep this member 8-byte aligned)
} PBUF;

typedef struct {						// for parsed command line args
	uint64_t fsize;						// parsed filesize (1st arg)
	enum { AF_FSIZE=1, AF_ARG=2, AF_FSTAR=4, AF_HEX=8 } flags; // flags
	PBUF *pPbuf0;						// ptr to 1st PBUF in list
	PBUF *pPbuf;						// ptr to current PBUF (last in list)
} ARGS;

void GetArgs(ARGS *pA);
void WriteData(ARGS *pA);

/*----------------------------------------------------------------------------
;
; Program entry point.  Note we can't use argc/argv -- we get the literal
; command line from the operating system in GetArgs().  This is because we
; need double quotes; can't have the standard library parsing them away.
;
;---------------------------------------------------------------------------*/

int __cdecl main(void) {
	ARGS a;

	GetArgs(&a);						// process command line
	WriteData(&a);						// write generated data to stdout
	return 0;
}

/*----------------------------------------------------------------------------
;
; Display help and exit.
;
;---------------------------------------------------------------------------*/

void HelpExit(void) {					// write help msg to stdout and exit
	fputs(
		"\nstdout ver 2.1. Generates byte pattern to stdout.\n\n"
		"  Usage:  stdout {-opts} {file-size{*}}"
				"  hex|\"ascii\"|@{file} ...\n\n"
		"Example:  stdout 100m ff 0y1010 \"ascii\\n\""
				" @file.txt 0n99 >100meg.tmp\n\n"
		"If 1st arg is ascii or @file, file-size is total argument size.\n"
		"See readme or paulhoule.com\\stdout for more options.\n"
	, stdout);
	exit(-1);
}

/*----------------------------------------------------------------------------
;
; Aborts w/error mes to stderr.
; fmt= printf format string, pstr= printf arg (optional, NULL if unused)
;
;---------------------------------------------------------------------------*/

void exiterr(char *fmt, char *pstr) {
	fprintf(stderr, "stdout: ");
	fprintf(stderr, fmt, pstr);
	exit(-1);
}

/*----------------------------------------------------------------------------
;
; Allocate memory; abort w/msg if allocation fails.
;
;---------------------------------------------------------------------------*/

void *mallocerr(size_t bytes) {
	void *pmem = malloc(bytes);
	if (pmem) return pmem;
	exiterr("Memory allocation error\n", NULL);
	return NULL;
}

/*----------------------------------------------------------------------------
;
; 32-bit *pRes = *pRes * mul + i.
; Returns upper 32-bit overflow (0 if none).
; Result and mul,i fields are 32-bit.
;
;---------------------------------------------------------------------------*/

uint32_t m32a(uint32_t *pRes, uint32_t mul, uint32_t i) {
	uint64_t t64 = (uint64_t) *pRes * mul + i;
	*pRes = (uint32_t) t64;
	return (uint32_t) (t64 >> 32);
}

/*----------------------------------------------------------------------------
;
; 64-bit *pRes = *pRes * mul + i.
; Returns upper 32-bit overflow (0 if none).
; Result and mul,i fields are 32-bit.
;
;---------------------------------------------------------------------------*/

uint32_t m64a(uint64_t *pRes, uint32_t mul, uint32_t i) {
	i = m32a((uint32_t *) pRes, mul, i);
	return m32a((uint32_t *) pRes + 1, mul, i);
}

/*----------------------------------------------------------------------------
;
; Convert ascii letter to lowercase.
;
;---------------------------------------------------------------------------*/

#define lower(c) ((c) | 0x20)

/*----------------------------------------------------------------------------
;
; Converts character to binary digit of any base (0..35).
; Result is >35 if input char not ascii 0-9,A-Z,a-z.
;
;---------------------------------------------------------------------------*/

unsigned GetDig(char c) {
	char d = c - '0';
	if ((unsigned char) d > 9) {
		d = lower(c) - 'a';
		if (d >= 0) d += 10;
	}
	return d;
}

/*----------------------------------------------------------------------------
;
; Convert string to unsigned 64-bit.
;
; Optional prefix 0x,0n,0t,0y (MSFT MASM convention) forces base 16,10,8,2.
;
; Optional suffix kmgtpe forces decimal kilo(x1000), mega, giga, tera, etc.
; Add i to suffix for power-of-two scaling (eg ki=x1024).
;
; Returns ptr to char stopping conversion.
;
;---------------------------------------------------------------------------*/

char *str2ull(
	uint64_t *pRes,						// &result, max 18446744073709551615
	char *pC,							// ptr to src
	unsigned base						// non-zero default base (eg 10, 16)
) {
	*pRes = 0;

	if (*pC == '0') {					// if potential 0x, 0y, etc..
		char *pb = "\x10x\xan\x8t\x2y";
		unsigned i;
		do { i = *pb++; } while (i && *pb++ != lower(pC[1]));
		if (i) {
			if (GetDig(pC[2]) < i) {	// valid digit must follow
				base = i, pC += 2;		// apply base, advance over prefix
			}
		}
	}

	do {								// process digits
		unsigned i = GetDig(*pC);
		if (i >= base) {				// if bad digit, check for postfix
			char sc;
			base = 0;
			i = 0;
			do {
				sc = "kmgtpe"[i++];
			} while (sc && sc != lower(*pC));
			if (!sc) break;
			if (lower(pC[1]) == 'i') base++;
			while (!m64a(pRes, (uint32_t) (1000 + base * 24), 0) && --i) ;
			if (!i) pC += base + 1;		// if no overflow, swallow postfix
			break;
		}
		if (m64a(pRes, (uint32_t) base, (uint32_t) i)) break;
	} while (*++pC);
	return pC;
}

/*----------------------------------------------------------------------------
;
; Allocate/Init empty buffer, and optionally append it to a list.
;
; Normally the default ->buf size is used (see definition of PBUF).
; A nonzero bsize (this is in bytes) overrides the default size.
;
; A total of the data bytes in all (if any) preceding blocks is placed in
; the newly allocated ->totprevbused field.
;
; The sum of ->totprevbused and ->bused in the active (last in list) block
; represents the total amount of data bytes in the entire block list.
;
; Note the difference between allocated and data bytes: ->totprevbused and
; ->bused deal with data bytes; ->bbytes is the allocated size of ->buf.
; No block anywhere in a list is required to be full.
;
; pPbuf= ptr to end (pPbuf->pNext == NULL) of a buffer list,
;		or NULL to not add new block to a list (eg starting a new list).
; bsize= if non-zero, overrides the size of the returned ->buf.
;
;---------------------------------------------------------------------------*/

PBUF *AllocPbuf(PBUF *pPbuf, unsigned bsize) {
	PBUF *pPtemp;

	if (!bsize) bsize = sizeof(pPtemp->buf); // set default
	pPtemp = mallocerr((sizeof(PBUF) - sizeof(pPtemp->buf)) + bsize);

	pPtemp->pNext = NULL;
	pPtemp->totprevbused = 0;
	pPtemp->bbytes = bsize;
	pPtemp->bused = 0;
	if (pPbuf) {
		pPbuf->pNext = pPtemp;
		pPtemp->totprevbused = pPbuf->totprevbused + pPbuf->bused;
	}
	return pPtemp;
}

/*----------------------------------------------------------------------------
;
; Appends data (which may be empty) to pattern buffer.
; pPbuf points to an existing PBUF (allowed to be completely full).
;
; Returns pPbuf, which may change if a new PBUF must be allocated.
;
;---------------------------------------------------------------------------*/

PBUF *ApptendToPbuf(PBUF *pPbuf, void *pData, unsigned n) {
	while (n) {
		unsigned nc = pPbuf->bbytes - pPbuf->bused; // buffer free bytes
		if (!nc) {						// if zero free in current buffer
			pPbuf = AllocPbuf(pPbuf, 0); // allocate new buffer
			nc = pPbuf->bbytes;
		}
		if (nc >= n) nc = n;			// limit copy to caller data size
		memcpy(pPbuf->buf + pPbuf->bused, pData, nc);
		pPbuf->bused += nc;
		(char *) pData += nc;
		n -= nc;
	}
	return pPbuf;
}

/*----------------------------------------------------------------------------
;
; Parses C-style string escape sequence \x{xxx...}
;
; In:	pT= ptr to char AFTER leading \
;		pO= ptr to where converted char should be stored
; Out:	pT advanced beyond processed escape sequence
;
;---------------------------------------------------------------------------*/

char *EscSeq(char *pO, char *pT) {
	unsigned char *pE =					// note: no Unicode (\u, \U) support
		"\2130" "\220x" "\15r" "\12n" "\134\\" "\42\"" "\47'"
		"\7a" "\10b" "\11t" "\13v" "\14f" "\33e" "\77?"
		"\2131" "\2132" "\2133" "\2134" "\2135" "\2136" "\2137";
	unsigned char c;
	unsigned char lc = *pT++;			// 2nd esc seq char (may be a '\0')

	do { c = *pE++; } while (c && *pE++ != lc); // look up 2nd char
	if (c) {							// if 2nd char of sequence matched,
		*pO = c;						// store (assuming c is result)
		if (!(c & 0x80)) return pT;		// if not hex or octal, we're done

		pE = pT;						// use pE to parse addtional chars
		*pO = 0;						// init output byte as accumulator
		if (c & 1) pE--;				// if octal, restore lead digit
		do {
			unsigned d = GetDig(*pE);	// try to get another digit
			unsigned b = c & 0x18;		// b= base operating in (16 or 8)
			if (d >= b) break;			// break= no more valid digits
			pE++;						// consume valid digit
			c |= 0x80;					// set error flag (assume overflow)
			d += *(unsigned char *)pO * b; // calc new sum of all digits
			*pO = (char) d;				// store new sum
			if (d > 255) break;			// break= overflow error
			c &= ~0x80;					// clear overflow/no-digit error flag
		} while (c & 16 || --c & 3);
		if (!(c & 0x80)) return pE;		// if no error, we're done
	}
	exiterr("Invalid escape sequence %s\n", pT - 2); // abort prog w/error
	return NULL;						// (prevent compiler error)
}

/*----------------------------------------------------------------------------
;
; Parses/stores string (eg "foo\n") given ptr to leading double quote.
;
; The string is appended to the PBUF pointed to by *ppPbuf.
; *ppPbuf may change if a new PBUF must be allocated.
;
; Returns advanced source ptr (beyond trailing double quote).
; If source ends with a '\0' (no trailing double quote exists), source ptr
; is left pointing to the trailing '\0'.
;
;---------------------------------------------------------------------------*/

char *GetStr(PBUF **ppPbuf, char *pT) {
	PBUF *pPbuf = *ppPbuf;				// output pattern buffer block

	pT++;								// advance over leading double quote
	while (1) {
		unsigned char c;
		if (pPbuf->bused == pPbuf->bbytes) { // if current block full,
			pPbuf = AllocPbuf(pPbuf, 0); // allocate new output block
		}
		c = *pT;
		if (!c) break;
		pT++;
		if (c == '"') break;
		{
			char *pO = pPbuf->buf + pPbuf->bused;
			*pO = c;
			if (c == '\\') pT = EscSeq(pO, pT); // process if esc sequence
		}
		pPbuf->bused++;
	}
	*ppPbuf = pPbuf;
	return pT;
}

/*----------------------------------------------------------------------------
;
; Return non-zero if *pSrc is a recognized (valid) argument terminator.
;
;---------------------------------------------------------------------------*/

int Term(char *pSrc) {
	int rc = 0;							// assume not a terminator
	uint32_t t = ' ' << 24 | '\t' << 16 | '"' << 8;	// t= ' ','\t','"','\0'

	do {								// check for terminator
		if (*pSrc == (char) t) rc++;
	} while (t >>= 8);

	return rc;
}

/*----------------------------------------------------------------------------
;
; Parses/stores hex number.
;
; The number is appended to the PBUF pointed to by *ppPbuf.
; *ppPbuf may change if a new PBUF must be allocated.
;
; If successful, returns advanced source ptr (beyond end-of-number), and
; number data (0..8 bytes) is appended to PBUF.
;
; If failure, returns entry pT value (nothing is stored).
;
;---------------------------------------------------------------------------*/

char *GetHex(PBUF **ppPbuf, char *pT) {
	uint64_t buf64;
	char *pE;

	pE = str2ull(&buf64, pT, 16);		// attempt to parse hex number
	if (pE != pT) {						// if char(s) parsed,
		unsigned i = 8;
		while (--i && !((char *) &buf64)[i]) ; // remove high order zeros
		i++;
		if (lower(*pE) == 'l') {		// if forcing length (0..8)
			i = (unsigned) (pE[1] - '0');
			if (i > 8) return pT;		// if > 8, bad length postfix
			pE += 2;
		}
		if (!Term(pE)) return pT;		// error if bad terminator
		*ppPbuf = ApptendToPbuf(*ppPbuf, (char *) &buf64, i);
	}
	return pE;
}

/*----------------------------------------------------------------------------
;
; Reads file content into pattern buffer.  fh= open file handle, or NULL
; to read from stdin.
;
; Returns zero if success.
; Returns non-zero if failure (some file data may have been added to buffer).
;
;---------------------------------------------------------------------------*/

int ReadAt(ARGS *pA, FILE *fh) {
	size_t lim;

	/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	; lim is an optimization -- the max file bytes we'll read.
	; If we've already buffered output filesize bytes, it makes no sense
	; to read more (perhaps gigabytes more) that won't be used.
	; (Note: careful... could have already buffered > output filesize.)
	*/

	lim = -1;							// assume we'll read max possible
	if (pA->flags & AF_FSIZE			// if total output size specified,
	 && !(pA->flags & AF_FSTAR)			//  and it's not "times pattern size"
	 && pA->fsize < lim) {				//	and it might limit what we'll read
		size_t tot =					// tot= total bytes buffered so far
			pA->pPbuf->totprevbused + pA->pPbuf->bused;
		lim = 0;						// assume no need to buffer more
		if (tot < (size_t) pA->fsize) {	// if haven't hit/exceeded fsize,
			lim = (size_t) pA->fsize - tot; // limit read to what we need
		}
	}

	if (fh == NULL) {					// handle NULL handle special case
		_setmode(_fileno(fh = stdin), _O_BINARY);
	}

	while (1) {							// loop to read file bytes
		size_t brd;						// bytes read (returned from fread)
		size_t brq =					// free bytes in current buffer (0-n)
		  pA->pPbuf->bbytes - pA->pPbuf->bused;
		if (!brq) {						// if buffer has no free space,
			brq = pA->pPbuf->bbytes;	// assume using previous alloc size
			if (brq < 64*1024*1024) brq *= 2; // double size to a point
			pA->pPbuf = AllocPbuf(pA->pPbuf, (unsigned) brq);
			brq = pA->pPbuf->bbytes;
		}
		if (brq > lim) brq = lim;		// enforce limit
		brd = fread(pA->pPbuf->buf + pA->pPbuf->bused, 1, brq, fh);
		pA->pPbuf->bused += (unsigned) brd;
		if (!(lim -= brd)) break;		// break= hit limit, we're done
		if (brd != brq) break;			// break= hit eof (or read error)
	}
	return ferror(fh);
}

/*----------------------------------------------------------------------------
;
; Processes @ argument -- this involves opening a file (if filename is
; specified, otherwise stdin is used) and reading the file/stream content
; into the pattern buffer.  Entry *pT must address lead arg '@' char.
;
; If successful, returns advanced source ptr (beyond @ argument).
; If argument ends with a '\0' returned ptr points to the '\0'.
;
; If failure, returns entry pT (some file data may be added to buffer).
;
;---------------------------------------------------------------------------*/

char *GetAt(ARGS *pA, char *pT) {
	char fn[260];						// used to hold path+filename
	char *pE = pT + 1;
	FILE *fh = NULL;					// NULL causes read from stdin

	do {								// do-nothing block (for breaks)
		char *p = fn;					// ptr used to store filename
		if (*pE == '"') pE++;			// if there, skip lead double quote
		while (1) {
			if (Term(pE)) {				// if (unquoted) terminator,
				if (pT[1] != '"') break; // terminate if not inside quotes
				if (!*pE) break;		// '\0' terminates inside quotes
				if (*pE == '"') {		// if quote while inside quotes,
					pE++;				// advance over closing quote
					break;				//	and terminate
				}
			}
			if (p == fn + sizeof(fn)) return pT; // abort if buffer overflow
			*p++ = *pE++;
		}
		*p = '\0';						// terminate filename

		if (p != fn) {					// if filename specified, open it
			fh = fopen(fn, "rb");
			if (!fh) break;				// break= error opening file
		}
		if (ReadAt(pA, fh)) break;		// read all file data, break if error

		pT = pE;						// advance pT to indicate success
	} while (0);

	if (fh) fclose(fh);					// close named file, if one opened
	return pT;
}

/*----------------------------------------------------------------------------
;
; Parses command line: gets output file size (if there) and creates
; pattern buffer.
;
; Note: even if the pattern generates no data (0 bytes), there is always 1
; buffer (PBUF) allocated -- it will be empty.
;
;---------------------------------------------------------------------------*/

__declspec(dllimport) char * __stdcall GetCommandLineA(void);

void GetArgs(ARGS *pA) {
	char *pCl, *pT;

	pA->flags = 0;						// init flags
	pA->pPbuf0 = pA->pPbuf = AllocPbuf(NULL, 0); // init pattern buffer list

	/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	; Init/advance beyond executable path/name in the command line.
	; There may be double quotes involved; Windows rules are used.
	*/

	pT = GetCommandLineA();				// asciiz command line we'll process
	while (*pT && *pT != ' ' && *pT != '\t') { // skip over program name
		if (*pT++ == '"') {				// skip double quoted string
			while (*pT && *pT++ != '"') ;
		}
	}

	pCl = NULL;							// need pCl = NULL at loop entry
	while (pT != pCl) {					// while something parsed w/o error,
		pCl = pT;						// advanced over parsed char(s)
		if (!*pT) break;				// break= hit end of command line

		if (*pT == ' ' || *pT == '\t') { // if whitespace
			pT++;						// swallow 1 whitespace char
			continue;
		}

		if (!(pA->flags & AF_ARG)		// if haven't seen an argument yet
		 && *pT >= '0' && *pT <= '9') {	//	and leading decimal digit

			char *pE = str2ull(&pA->fsize, pT, 10); // parse as decimal
			if (*pE == '*') *pE++, pA->flags |= AF_FSTAR; // handle '*'
			if (!Term(pE)) break;		// error if bad terminator
			pT = pE;
			pA->flags |= AF_ARG | AF_FSIZE; // flag arg and fsize seen
			continue;
		}

		if (*pT == '"') {
			pT = GetStr(&pA->pPbuf, pT); // process double quoted string
		} else if (*pT == '@') {
			pT = GetAt(pA, pT);			// process file data
		} else if (*pT == '-' && lower(pT[1]) == 'h') {
			pT += 2, pA->flags |= AF_HEX; // handle -h switch
			continue;
		} else {
			/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
			; Here arg can only be a hex (by default) number, but only
			; if a previous arg has been seen.
			*/
			if (pA->flags & AF_ARG) pT = GetHex(&pA->pPbuf, pT);
		}
		pA->flags |= AF_ARG;			// flag an argument has been seen
	}
	if (*pCl) exiterr("Argument error %s\n", pCl);
	if (!(pA->flags & AF_ARG)) HelpExit(); // give help if no arguments
}

/*----------------------------------------------------------------------------
;
; Writes the parsed output pattern to stdout.
;
;---------------------------------------------------------------------------*/

void WriteData(ARGS *pA) {
	PBUF *pPbuf = pA->pPbuf;			// last in pattern buffer block list
	size_t tdata =						// total pattern buffer data bytes
		pPbuf->totprevbused + pPbuf->bused;

	/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	; Generate final output filesize (pA->fsize).  It may be fine already,
	; or it may not have been given and we have to use the pattern size, or
	; if file-size* used we have to multiply the pattern size into it.
	*/

	if (pA->flags & AF_FSIZE) {			// if output filesize was specified,
		if (pA->flags & AF_FSTAR) {		// if * postfix used,
			uint64_t fsize = pA->fsize;
			pA->fsize = tdata;			// set up to use m64a
			if (fsize != (uint32_t) fsize // if multiply overflow,
			 || m64a(&pA->fsize, (uint32_t) fsize, 0)) {
				exiterr("file-size* too big\n", NULL);
			}
		}
	} else {							// else filesize not specified,
		pA->fsize = tdata;				// use size of pattern data
	}

	/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	; If pattern buffer is empty, put a zero -- the default output value --
	; in it (we need some pattern data to replicate).
	*/

	if (!tdata) {						// if no data in pattern buffer,
		pPbuf->buf[0] = '\0';			// add a zero (the default data value)
		pPbuf->bused++;
	}

	/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	; If just one pattern block that's less than half full, we can replicate
	; data in the block to enable larger fwrite sizes (don't want to fwrite
	; 1 byte a billion times).  We try to get the block close to full.
	*/

	if (!pPbuf->totprevbused) {			// if all data is in the pPbuf block
		uint32_t n = pPbuf->bused;		// n = # bytes copied to extend buf
		while (pPbuf->bused + n <= sizeof(pPbuf->buf)
		 && pPbuf->bused < pA->fsize) {	// replicate while it makes sense to
			memcpy(pPbuf->buf + pPbuf->bused, pPbuf->buf, n);
			pPbuf->bused += n;			// show block data has increased
			if (n < 120) n <<= 1;		// increase copy size to a point
		}
	}

	/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	; Write data to stdout.  First we link the last pattern data block to
	; the first, creating a circular list.  That way we can just keep moving
	; to the "next" block until all the data we need to generate is written.
	*/

	_setmode(_fileno(stdout), _O_BINARY); // must write in binary mode

	pPbuf->pNext = pA->pPbuf0;			// create circular list
	while (1) {
		uint32_t bused;

		pPbuf = pPbuf->pNext;			// move to next pattern buffer
		bused = pPbuf->bused;			// data bytes in this ->buf
		if (!bused) continue;			// ignore (it's possible!) empty block

		if (bused > pA->fsize) {		// limit to what's left to write
			bused = (uint32_t) pA->fsize;
			if (!bused) break;			// break= no more data to write
		}
		pA->fsize -= bused;				// update what's left to write

		if (pA->flags & AF_HEX) {		// if writing in hex,
			uint8_t *p = pPbuf->buf;	// dump formatted hex bytes to stdout
			do { printf("%02x ", *p++); } while (--bused);
			continue;
		}
		if (bused != fwrite(pPbuf->buf, 1, bused, stdout)) {
			exiterr("*** File write error ***\n", NULL);
		}
	}
}
